jQuery(document).ready(function ($) {
    'use strict';

    /* Zustand Vanilla */
    function create(createState) {
        var state;
        var listeners = new Set();

        const setState = (partial, replace) => {
            var nextState = typeof partial === 'function' ? partial(state) : partial;

            if (nextState !== state) {
                var _previousState = state;
                state = replace ? nextState : Object.assign({}, state, nextState);
                listeners.forEach(function (listener) {
                    return listener(state, _previousState);
                });
            }
        };

        const getState = () => state;

        var subscribeWithSelector = function (listener, selector, equalityFn) {
            if (selector === void 0) {
                selector = getState;
            }

            if (equalityFn === void 0) {
                equalityFn = Object.is;
            }

            var currentSlice = selector(state);

            function listenerToAdd() {
                var nextSlice = selector(state);

                if (!equalityFn(currentSlice, nextSlice)) {
                    var _previousSlice;
                    if (typeof currentSlice === 'object') {
                        _previousSlice = Array.isArray(currentSlice) ? [...currentSlice] : {...currentSlice};
                    } else {
                        _previousSlice = currentSlice;
                    }

                    listener(currentSlice = nextSlice, _previousSlice);
                }
            }

            listeners.add(listenerToAdd);
            return function () {
                return listeners.delete(listenerToAdd);
            };
        };

        var subscribe = function (listener, selector, equalityFn) {
            if (selector || equalityFn) {
                return subscribeWithSelector(listener, selector, equalityFn);
            }

            listeners.add(listener);
            return function () {
                return listeners.delete(listener);
            };
        };

        const destroy = () => listeners.clear();

        var api = {setState, getState, subscribe, destroy};

        state = createState(setState, getState, api);

        return api;
    }

    /* App **********************************************/

    const editStore = create((set, get, api) => ({
        view: '',
        textEditing: '',
        imageEditing: '',
        width: 0,
        height: 0,
        margin: {},
        images: [],
        texts: [],

        init: (data) => set((state) => ({...state, ...data})),

        setTextActive: (textId = '') => set((state) => ({...state, textEditing: textId})),

        setWidth: (width) => set((state) => ({...state, width: +width})),

        setHeight: (height) => set((state) => ({...state, height: +height})),

        setMargin: (pos, value) => set((state) => {
            let tmpMargin = {...state.margin};
            tmpMargin[pos] = +value;
            state.margin = tmpMargin;
            return {...state};
        }),

        setAllMargin: (value) => set((state) => ({...state, margin: value})),

        // setImage: (id, url) => {
        //     return set((state) => {
        //         let {view} = state,
        //             currentView = {...state[view]};
        //
        //         currentView.imgId = id;
        //         currentView.imgUrl = url;
        //         state[view] = currentView;
        //
        //         return {...state};
        //     })
        // },

        setImage: (id, url, title) => {
            return set((state) => {
                let {images} = state;
                images.push({id, url, title, top: 0, left: 0});

                return {...state, images: [...images]};
            })
        },

        removeImage: (imgId) => {
            set((state) => {
                let {images} = state,
                    currentImgId = images.findIndex(image => image.id === imgId);

                if (currentImgId === -1) return state;

                images.splice(currentImgId, 1);

                return {...state, images: [...images]};
            })
        },

        setBgColor: (color) => {
            return set((state) => {
                let {view} = state,
                    currentView = {...state[view]};

                currentView.background_color = color;
                state[view] = currentView;

                return {...state};
            })
        },

        addText: () => {
            set((state) => {
                let {texts} = state;

                if (!texts) texts = [];

                texts.push({
                    id: Date.now(),
                    content: 'Text',
                    pos_x: 0,
                    pos_y: 0,
                    color: '#000000',
                    size: 16,
                    font_family: 'arial',
                    font_weight: 'normal'
                });

                return {...state, texts: [...texts]};
            })
        },

        updateText: (key, value) => {
            set((state) => {
                let {texts, textEditing} = state,
                    currentTextId = texts.findIndex(text => text.id === textEditing);

                if (currentTextId === -1) return state;

                let currentText = texts[currentTextId];

                currentText[key] = value;

                texts[currentTextId] = currentText;

                return {...state, texts: [...texts]};
            })
        },

        updateTextPosition: (id, x, y) => {
            set((state) => {
                let {texts} = state,
                    currentTextId = texts.findIndex(text => text.id === id);

                if (currentTextId === -1) return state;

                let currentText = texts[currentTextId];

                currentText.pos_x = x;
                currentText.pos_y = y;

                texts[currentTextId] = currentText;

                return {...state, texts: [...texts]};
            })
        },

        updateImagePosition: (id, params = {}) => {
            set((state) => {
                let {x = null, y = null, w = null} = params,
                    {images} = state,
                    currentImgId = images.findIndex(img => img.id === id);

                if (currentImgId === -1) return state;

                let currentImg = images[currentImgId];

                if (x !== null) currentImg.pos_x = x;
                if (y !== null) currentImg.pos_y = y;
                if (w !== null) currentImg.width = w;

                images[currentImgId] = currentImg;

                return {...state, images: [...images]};
            })
        },

        getText: (textId) => {
            let state = get();

            if (!textId) return null;

            let {texts} = state;

            let currentTextId = texts.findIndex(text => text.id === textId);

            if (currentTextId !== -1) {
                return texts[currentTextId];
            }

            return null;
        },

        removeText: (textId) => {
            set((state) => {
                let {texts} = state,
                    currentTextId = texts.findIndex(text => text.id === textId);

                if (currentTextId === -1) return state;

                texts.splice(currentTextId, 1);

                return {...state, texts: [...texts]};
            })
        }
    }));

    /* GLOBAL PROPERTIES *******************************/

    function ImageControlList() {
        let self = this;
        this.el = $('#gift4u-image-list');
        this.media = wp.media({multiple: false});

        this.addMediaListener = function () {
            $('.gift4u-add-image').on('click', function () {
                self.media.open().off('select').on('select', function (e) {
                    let uploadedImages = self.media.state().get('selection').first();
                    let selectedImages = uploadedImages.toJSON();
                    let {id, url, title} = selectedImages;
                    editStore.getState().setImage(id, url, title);
                })
            });
        };

        this.render = function (images) {
            let imgHtml = images.map(img => {
                return `<div class="gift4u-image-control" data-image_id="${img.id}"><span class="gift4u-image-title">${img.title}</span><span class="gift4u-remove-image">×</span></div>`
            }).join('');

            self.el.html(imgHtml);
        };

        this.setImgActive = function () {

        };

        this.removeImage = function () {
            let imgId = $(this).closest('.gift4u-image-control').data('image_id');
            editStore.getState().removeImage(imgId)
        };

        this.addMediaListener();
        this.render(editStore.getState().images);

        /*Events*/
        this.el.on('click', '.gift4u-image-control', this.setImgActive)
            .on('click', '.gift4u-remove-image', this.removeImage);

        editStore.subscribe(this.render, s => s.images);
    }

    function TextControlList() {
        let self = this;
        this.el = $('#gift4u-text-list');

        this.addTextListener = function () {
            $('.gift4u-add-text').on('click', function () {
                editStore.getState().addText();
            });
        };

        this.render = function (texts) {
            let {textEditing} = editStore.getState();

            let imgHtml = texts.map(text => {
                let {id, content = ''} = text;

                if (!content) content = id;

                let active = textEditing === id ? 'gift4u-text-control-active' : '';

                return `<div class="gift4u-text-control ${active}" data-text_id="${id}">
                            <span class="gift4u-text-control-content">${content}</span>
                            <span class="gift4u-remove-text">×</span>
                        </div>`;

            }).join('');

            self.el.html(imgHtml);
        };

        this.setTextActive = function () {
            const {setTextActive} = editStore.getState();
            setTextActive($(this).data('text_id'));
        };

        this.removeText = function () {
            let textId = $(this).closest('.gift4u-text-control').data('text_id');
            editStore.getState().removeText(textId)
        };

        this.hightlightRow = function (textId) {
            self.el.find('.gift4u-text-control').removeClass('gift4u-text-control-active');
            self.el.find(`.gift4u-text-control[data-text_id=${textId}]`).addClass('gift4u-text-control-active');
        };

        this.addTextListener();
        this.render(editStore.getState().texts);

        /*Events*/
        this.el.on('click', '.gift4u-text-control', this.setTextActive)
            .on('click', '.gift4u-remove-text', this.removeText);

        editStore.subscribe(this.render, s => ([...s.texts]));
        // editStore.subscribe(this.hightlightRow, s => s.textEditing);
    }

    function GlobalControl() {
        this.el = $('#gift4u-global-properties');

        this.setWidth = function () {
            editStore.getState().setWidth($(this).val());
        };

        this.setHeight = function () {
            editStore.getState().setHeight($(this).val());
        };

        this.setWidthHeight = function () {
            let size = $(this).val();
            switch (size) {
                case 'a4':
                    editStore.getState().setWidth(210);
                    editStore.getState().setHeight(297);
                    break;

                case 'a3':
                    editStore.getState().setWidth(297);
                    editStore.getState().setHeight(420);
                    break;

                case 'a5':
                    editStore.getState().setWidth(148.5);
                    editStore.getState().setHeight(210);
                    break;
            }
        };

        this.setMargin = function () {
            let pos = $(this).data('position');
            editStore.getState().setMargin(pos, $(this).val());
        };

        this.setPaperSize = function () {
            let size = 'custom',
                {width, height} = editStore.getState();

            switch (true) {
                case width === 210 && height === 297:
                    size = 'a4';
                    break;

                case width === 297 && height === 420:
                    size = 'a3';
                    break;

                case width === 148.5 && height === 210:
                    size = 'a5';
                    break;
            }

            $('.gift4u-paper-size').val(size);
        };

        this.el.on('change', '.gift4u-card-view-width', this.setWidth)
            .on('change', '.gift4u-card-view-height', this.setHeight)
            .on('change', '.gift4u-card-view-margin', this.setMargin)
            .on('change', '.gift4u-paper-size', this.setWidthHeight);

        let actions = {
            width: (state) => {
                $('.gift4u-card-view-width').val(state);
                this.setPaperSize();
            },
            height: (state) => $('.gift4u-card-view-height').val(state),
            margin: (m) => {
                for (let pos in m) {
                    $(`.gift4u-card-view-margin[data-position=${pos}]`).val(m[pos]);
                }
            }
        };

        for (let key in actions) {
            editStore.subscribe(actions[key], state => state[key]);
        }

        new ImageControlList();
        new TextControlList();
    }


    /* CARD VIEW *******************************/
    function View() {
        let self = this;
        this.paper = $(`.gift4u-card-paper`);
        this.cardContent = $(`.gift4u-card-content`);

        this.renderImages = function (images, prevState) {
            let removeArray = prevState.filter(x => {
                let index = images.findIndex(image => image.id === x.id);
                return index === -1;
            });

            if (removeArray.length) {
                for (let el of removeArray) {
                    $(`.gift4u-image[data-image_id=${el.id}]`).remove();
                }
            }

            if (!images.length) return;

            for (let image of images) {
                let check = $(`.gift4u-image[data-image_id=${image.id}]`);

                if (check.length) {
                    check.css({top: image.pos_y, left: image.pos_x});
                } else {
                    let imgHtml = $(`<div class="gift4u-image" data-image_id="${image.id}"><img src="${image.url}" /></div>`);

                    imgHtml.css({top: image.pos_y, left: image.pos_x, position: 'absolute', width: image.width});

                    imgHtml.draggable({
                        containment: 'parent',
                        stop: function (event, ui) {
                            let {top, left} = ui.position,
                                imgId = ui.helper.data('image_id');
                            editStore.getState().updateImagePosition(imgId, {x: left, y: top});
                        }
                    });

                    self.cardContent.append(imgHtml);

                    imgHtml.resizable({
                        containment: ".gift4u-card-content",
                        aspectRatio: true,
                        handles: "se",
                        stop: function (event, ui) {
                            let {width} = ui.size,
                                imgId = ui.helper.data('image_id');
                            editStore.getState().updateImagePosition(imgId, {w: width});
                        }
                    })
                }
            }
        };

        this.renderTexts = function (texts, prevState) {
            let removeArray = prevState.filter(x => {
                let index = texts.findIndex(text => text.id === x.id);
                return index === -1;
            });

            if (removeArray.length) {
                for (let el of removeArray) {
                    $(`.gift4u-text[data-text_id=${el.id}]`).remove();
                }
            }

            if (!texts.length) return;


            for (let text of texts) {
                let check = $(`.gift4u-text[data-text_id=${text.id}]`);
                let textContent = text.content,
                    width = text.width !== '' ? `${text.width}mm` : 'auto',
                    css = {
                        top: +text.pos_y,
                        left: +text.pos_x,
                        fontSize: `${text.size}px`,
                        fontWeight: text.font_weight,
                        fontFamily: text.font_family,
                        color: text.color,
                        width: width,
                        textAlign: text.align,
                    };

                for (let sc in gift4uParams.replaceShortcodes) {
                    textContent = textContent.replaceAll(sc, gift4uParams.replaceShortcodes[sc]);
                }
                if (check.length) {
                    check.css(css);
                    check.html(textContent)
                } else {
                    let textHtml = $(`<span class="gift4u-text" data-text_id="${text.id}">${textContent}</span>`);

                    textHtml.css({...css, position: 'absolute', zIndex: 99});

                    textHtml.draggable({
                        containment: 'parent',
                        stop: function (event, ui) {
                            let {top, left} = ui.position, textId = ui.helper.data('text_id');
                            editStore.getState().updateTextPosition(textId, left, top);
                            $('.gift4u-text-pos-x').val(left);
                            $('.gift4u-text-pos-y').val(top);
                        }
                    });

                    self.cardContent.append(textHtml);
                }
            }

            self.cardContent.find('.gift4u-text').removeClass('gift4u-text-view-active');
            let textEditing = editStore.getState().textEditing;
            if (textEditing) self.cardContent.find(`.gift4u-text[data-text_id=${textEditing}]`).addClass('gift4u-text-view-active');
        };

        this.cardContent.on('click', '.gift4u-text', function () {
            const {setTextActive} = editStore.getState();
            setTextActive($(this).data('text_id'));
        });

        let actions = {
            width: (state) => this.paper.css('width', state + 'mm'),
            height: (state) => this.paper.css('height', state + 'mm'),
            margin: (m) => this.paper.css({paddingLeft: `${m.left}mm`, paddingTop: `${m.top}mm`, paddingRight: `${m.right}mm`, paddingBottom: `${m.bottom}mm`,}),
        };

        for (let key in actions) {
            editStore.subscribe(actions[key], s => s[key]);
        }

        editStore.subscribe(this.renderTexts, s => [...s.texts]);
        editStore.subscribe(this.renderImages, s => [...s.images]);

    }

    function Text(text = {}) {
        let self = this;

        this.onChange = (e, type) => {
            const {updateText} = editStore.getState();
            let value = e.target.value;
            updateText(type, value);
        };

        this.onTextAlignChange = (e) => {
            let thisBtn = $(e.currentTarget);
            let value = thisBtn.val();
            const {updateText} = editStore.getState();
            thisBtn.closest('.gift4u-text-align').find('button').removeClass('active');
            updateText('align', value);
            thisBtn.addClass('active');
        };

        this.onColorChange = (value) => {
            const {updateText} = editStore.getState();
            updateText('color', value);
        };

        this.close = () => {
            const {setTextActive} = editStore.getState();
            setTextActive();
        };

        this.shortcodeAction = function () {
            $(this).find('.gift4u-text-shortcode-list').toggle();
        };

        this.addShortcode = function () {
            let sc = $(this).html(),
                pos = self.textPos,
                textInput = self.el.find('.gift4u-text-content'),
                currentText = textInput.val(),
                newText = '';

            if (pos) {
                let before = currentText.substr(0, pos);
                let after = currentText.substr(pos);
                newText = before + sc + after;
            } else {
                newText = currentText + sc;
            }

            textInput.val(newText).trigger('keyup');
        };

        this.getPosition = function () {
            self.textPos = this.selectionStart;
        };

        this.render = function () {
            let {content = '', pos_x = 0, pos_y = 0, size = 16, font_family = '', color = '#000000', font_weight = 'normal', width = '', align = 'left'} = text;

            let fontFamilyOptions = gift4uParams.fontFamilies.map((f) => {
                let selected = f.value === font_family ? 'selected' : '';
                return `<option value="${f.value}" ${selected}>${f.text}</option>`;
            }).join('');

            let fontWeightOptions = gift4uParams.fontWeight.map((f) => {
                let selected = f === font_weight ? 'selected' : '';
                return `<option value="${f.value}" ${selected}>${f.text}</option>`;
            }).join('');

            let html = $(`<div class="gift4u-text-panel-inner">
                            <h4 class="vi-ui header">Text edit<i class="vi-ui icon x gift4u-close-edit-text"> </i></h4>
                            <div class="fields">
                                <div class="field">
                                    <label>Pos X</label>
                                    <input type="number" class="gift4u-text-pos-x" min="0" value="${pos_x}">
                                </div>
                                <div class="field">
                                    <label>Pos Y</label>
                                    <input type="number" class="gift4u-text-pos-y" min="0" value="${pos_y}">
                                </div>
                            </div>
                            <div class="field">
                                <label>Content</label>
                                <div class="vi-ui right labeled input">
                                  <input type="text" class="gift4u-text-content" value="${content}">
                                  <div class="vi-ui basic label gift4u-text-content-shortcode">
                                    {s}
                                    <div class="gift4u-text-shortcode-list"> 
                                        <div class="gift4u-text-shortcode">{site_title}</div>
                                        <div class="gift4u-text-shortcode">{site_url}</div>
                                        <div class="gift4u-text-shortcode">{sender_name}</div>
                                        <div class="gift4u-text-shortcode">{recipient_name}</div>
                                        <div class="gift4u-text-shortcode">{message}</div>
                                        <div class="gift4u-text-shortcode">{coupon_code}</div>
                                        <div class="gift4u-text-shortcode">{coupon_value}</div>
                                        <div class="gift4u-text-shortcode">{coupon_expiry_date}</div>
                                    </div>
                                  </div>
                                </div>
                            </div>               
                            <div class="field">
                                <label>Width</label>
                                <div class="vi-ui right labeled input">
                                  <input type="number" class="gift4u-text-width" value="${width}" placeholder="Default: auto" min="0">
                                  <div class="vi-ui basic label ">mm</div>
                                </div>
                            </div>  
                            <div class="field">
                                <label>Align</label>
                                <div class="vi-ui icon basic small buttons gift4u-text-align">
                                  <button type="button" class="vi-ui button" value="left"><i class="align left icon"> </i></button>
                                  <button type="button" class="vi-ui button" value="center"><i class="align center icon"> </i></button>
                                  <button type="button" class="vi-ui button" value="right"><i class="align right icon"> </i></button>
                                </div>
                            </div>
                            <div class="field">
                                <label>Color</label>
                                <input type="text" class="gift4u-text-color" value="${color}">
                            </div>
                            <div class="field">
                                <label>Font size</label>
                                <input type="number" class="gift4u-text-size" min="0" value="${size}">
                            </div>
                            <div class="field">
                                <label>Font weight</label>
                                <select class="gift4u-text-font-weight">
                                    ${fontWeightOptions}
                                </select>
                            </div>
                        </div>`);

            //                    <div className="field">
            //                                 <label>Font</label>
            //                                 <select className="gift4u-text-font-family">
            //                                     ${fontFamilyOptions}
            //                                 </select>
            //                             </div>

            html.find(`.gift4u-text-align button[value=${align}]`).addClass('active');

            html.on('focusout', '.gift4u-text-content', this.getPosition)
                .on('click', '.gift4u-text-content-shortcode', this.shortcodeAction)
                .on('click', '.gift4u-text-shortcode', this.addShortcode)
                .on('keyup', '.gift4u-text-content', (e) => this.onChange(e, 'content'))
                .on('change keyup', '.gift4u-text-width', (e) => this.onChange(e, 'width'))
                .on('change keyup', '.gift4u-text-pos-x', (e) => this.onChange(e, 'pos_x'))
                .on('change keyup', '.gift4u-text-pos-y', (e) => this.onChange(e, 'pos_y'))
                .on('change keyup', '.gift4u-text-size', (e) => this.onChange(e, 'size'))
                .on('change', '.gift4u-text-font-family', (e) => this.onChange(e, 'font_family'))
                .on('change', '.gift4u-text-font-weight', (e) => this.onChange(e, 'font_weight'))
                .on('click', '.gift4u-text-align button', this.onTextAlignChange)
                .on('click', '.gift4u-close-edit-text', this.close);

            html.find('.gift4u-text-color').colorPicker({
                renderCallback($elm, toggled) {
                    self.onColorChange($elm[0].value);
                }
            });

            self.el = html;

            return html;
        };

        return this.render();
    }

    function TextControl() {
        this.el = $('#gift4u-text-control-panel');

        this.render = (textId = '') => {
            if (!textId) {
                this.el.hide();
                return;
            }

            const {getText} = editStore.getState();
            let selectedText = getText(textId);

            if (!selectedText) return;

            this.el.html(new Text(selectedText)).fadeIn(200);
            this.el.show();
            let textContentField = this.el.find('.gift4u-text-content');
            let strLength = textContentField.val().length * 2;// Opera sometimes sees a carriage return as 2 characters.
            textContentField.trigger('focus');
            textContentField[0].setSelectionRange(strLength, strLength);
        };

        editStore.subscribe(this.render, s => s.textEditing);
    }

    function App() {
        function resetView() {
            const {setTextActive} = editStore.getState();
            setTextActive();
        }

        $('body')
            .on('click', function (e) {
                let thisTarget = $(e.target);

                if (thisTarget.find('#gift4u-card-editor').length > 0) resetView();

                if (!thisTarget.closest('.gift4u-text-content-shortcode').length && !thisTarget.is('.gift4u-text-content-shortcode')) {
                    $('.gift4u-text-shortcode-list').hide();
                }
            });

        $('form#post').on('submit', function (e, v) {
            const {width, height, margin, images, texts} = editStore.getState();
            let data = JSON.stringify({width, height, margin, images, texts});
            console.log(data);

            let inputData = `<input type="hidden" name="gift4u_card_template_data" value='${data}'/>`;
            $(this).append(inputData);

            return true;
        });

        let initTemplate;
        if (typeof gift4uParams.cardTemplate === 'object') {
            initTemplate = gift4uParams.cardTemplate;
        } else {
            initTemplate = {width: 210, height: 297, margin: {top: 15, left: 15, right: 15, bottom: 15}};
        }

        editStore.getState().init(initTemplate);
    }

    new GlobalControl();
    new TextControl();
    new View();
    App();

});